/* 
	Author: Iosif Adrian Mihai
	Inspired by PM labs - UPB ACS
	http://cs.curs.pub.ro/wiki/pm/
	
	4 * 4 cube lighting program:
	implements a brownian-like random lighting
	and a linear movement, both influenced by
	exterior lighting given through sensor
	pressing button(PD6) changes lighting mod/params
*/
#include <avr/io.h>
#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>


/*globals used for cube lighting state*/
int leds[16] = {PB0, PB1, PB2, PB3, PB4, PB5, PB6, PB7, PA7, PA6, PA5, PA4, PC3, PC2, PC1, PC0};

int levels[4] = {PD5, PD4, PD1, PD2};

int state_delays[8] = {0, 200, 0, 0, 0, 30, 60, 90};


int lvls[8] = {4, 4, 2, 1, 0, 0, 0, 0};


int lds[8] = {16, 16, 2, 1, 0, 0, 0, 0};

int state = 0;

int max_levels = 4;
int max_leds = 16;

/***********************************************/

/* get sensor value */
int ADC_get(uint8_t channel)
{

	ADMUX = (ADMUX & ~(0x1f << MUX0)) | channel;

	ADCSRA |= (1 << ADSC);
	while(ADCSRA & (1 << ADSC));

	return ADC;
}

/* cube lighting that simulates a brownian movement */
void brownian_move() {
	int i = 0;
	int j = 0;
	int nr_levels;
	int nr_leds;
	
	int marked_levels[4];
	int marked_leds[16];
	//no leds/levels marked
	for (i = 0; i < 4; ++i) {
		marked_levels[i] = 0;
	}
	
	for (i = 0; i < 16; ++i) {
		marked_leds[i] = 0;
	}
	
	/* choose random num of levels and leds, within 
	current state constraints */
	nr_levels = 1 + rand() % 4;
	while(nr_levels > max_levels)
		nr_levels = 1 + rand() % 4;
	nr_leds = 1 + rand() % 16;
	while(nr_leds > max_leds)
		nr_leds = 1 + rand() % 16;
	while(nr_levels--) {
		int lev = rand() % 4;
		while(marked_levels[lev] == 1) {
			lev = rand() % 4;
		}
		marked_levels[lev] = 1;
	}
	
	while(nr_leds--) {
		int led = rand() % 16;
		while(marked_leds[led] == 1) {
			led = rand() % 16;
		}
		marked_leds[led] = 1;
	}
	/* init registers */
	DDRB = 0;
	DDRC = 0;
	DDRA = 0;
	DDRD = 0;
	PORTB = 0;
	PORTC = 0;
	PORTA = 0;
	PORTD = 0;
	DDRD = 0;
	DDRD &= ~(1<<PD6);
	PORTD = 0;
	PORTD |= (1<<PD6);
	
	/* light up chosen levels and leds */
	for (i = 0; i < 4; ++i) {
		if (marked_levels[i] == 1) {
			DDRD |= (1<<levels[i]);
			PORTD |= (1<<levels[i]);
		}
	}
	
	for (j = 0; j < 16; ++j) {
		if (j < 8 && marked_leds[j]) {
			DDRB |= (1 << leds[j]);
			PORTB |= (1 << leds[j]);	
		} else if (j >= 8 && j < 12 && marked_leds[j]) {							
			DDRA |= (1 << leds[j]);
			PORTA |= (1 << leds[j]);
					
		} else if (marked_leds[j]){
			DDRC |= (1 << leds[j]);
			PORTC |= (1 << leds[j]);
		}
	}
}


/* delay for brownian lights */
void brownian_delay(double current, double median, int delay_overhead)
{
	if (current < median / 4) {
		_delay_ms(1500 + delay_overhead);
	} else if (current >= median / 4 && current < median / 2) {
		_delay_ms(1000 + delay_overhead);
	} else if (current >= median / 2  && current < median) {
		_delay_ms(500 + delay_overhead);
	} else if (current >= median  && current < median * 2) {
		_delay_ms(250 + delay_overhead);
	} else if (current >= median * 2 && current < median * 3) {
		_delay_ms(100 + delay_overhead);
	} else if (current >= median * 3) {
		_delay_ms(50 + delay_overhead);
	}
}


/* delay for linear lights */
int linear_delay(double current, double median, int delay_overhead)
{
	int cont_delay = 0;
	if (current < median / 4) {
		cont_delay = 250;
	} else if (current >= median / 4  && current < median / 2) {
		cont_delay = 200;
	} else if (current >= median / 2  && current < median) {
		cont_delay = 150;
	} else if (current >= median  && current < median * 2) {
		cont_delay = 100;
	} else if (current >= median * 2 && current < median * 3) {
		cont_delay = 50;
	} else if (current >= median * 3) {
		cont_delay = 10;
	}
	cont_delay += delay_overhead;
	return cont_delay;

}
/* check if PD6 is pressed and change state accordingly
	calculate median again
*/
void check_button(int state, int *new_state, int *pressed, double *median)
{
	int k;
	while ((!(PIND & (1 << PD6)))) {
		*new_state = (state + 1) % 8;
		*pressed = 1;
		_delay_ms(20);
		/* recalc median */
		for (k = 0; k < 30; ++k) {
			*median += ADC_get(0);
		}
		*median = *median / 30;	
	}
}

/* linear movement lighting*/
void linear_move(int *state, int *new_state, int *pressed, int delay_overhead, double *median)
{
	int i,j;
	int cont_delay;
	/* init register */
	DDRB = 0;
	DDRC = 0;
	DDRA = 0;
	DDRD = 0;
	PORTB = 0;
	PORTC = 0;
	PORTA = 0;
	PORTD = 0;
	DDRD = 0;
	DDRD &= ~(1<<PD6);
	PORTD = 0;
	PORTD |= (1<<PD6);
	
	for (i = 0; i < 4; ++i) {
		if (*pressed == 1) break;
		for (j = 0; j < 16; ++j) {
			if (*pressed == 1) break;
			double current = ADC_get(0);
			cont_delay = linear_delay(current, *median, delay_overhead);
			/* reinit registers */
			DDRB = 0;
			DDRC = 0;
			DDRA = 0;
			DDRD = 0;
			PORTB = 0;
			PORTC = 0;
			PORTA = 0;
			PORTD = 0;
			DDRD = 0;
			DDRD &= ~(1<<PD6);
			PORTD = 0;
			PORTD |= (1<<PD6);
			/* light leds */
			if (j < 8 ) {
				DDRD |= (1<<levels[i]);
				PORTD |= (1<<levels[i]);
				DDRB |= (1 << leds[j]);
				PORTB |= (1 << leds[j]);	
			} else if (j >= 8 && j < 12) {
				DDRD |= (1<<levels[i]);
				PORTD |= (1<<levels[i]);							
				DDRA |= (1 << leds[j]);
				PORTA |= (1 << leds[j]);	
			} else {
				DDRD |= (1<<levels[i]);
				PORTD |= (1<<levels[i]);
				DDRC |= (1 << leds[j]);
				PORTC |= (1 << leds[j]);
			}
			
			/* add delay */
			_delay_ms(cont_delay);
			
			*new_state = *state;
			check_button(*state, new_state, pressed, median);
			*state = *new_state;
		}
	}
}



int main(void) {
	int pressed = 0;
	int i;
	int new_state;
	double median = 0;
	int delay_overhead = 0;
	/*  set up ADC */
	ADMUX = (1 << REFS0);
	ADCSRA = (1 << ADEN) | (7 << ADPS0);
	/* calc initial light median */
	for (i = 0; i < 30; ++i) {
		median += ADC_get(0);
	}
	median = median / 30; /* computed median temperature */


	while (1) {
		pressed = 0;
		delay_overhead = state_delays[state];
		max_leds = lds[state];
		max_levels = lvls[state];
		if (state < 4) {
			brownian_move();
			double current = ADC_get(0);
			brownian_delay(current, median, delay_overhead);
			new_state = state;
			check_button(state, &new_state, &pressed, &median);
			state = new_state;
		} else {
			linear_move(&state, &new_state, &pressed, delay_overhead, &median);
		}
	}
	
	return 0;
}

